/** 
 *   xml_proc: read, proces and write xml data file. 
 *
 *   (c) Copyright 2010 Federico Bareilles <bareilles@gmail.com>,
 *
 *   This program is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU General Public License as
 *   published by the Free Software Foundation; either version 2 of
 *   the License, or (at your option) any later version.
 *     
 *   The author does NOT admit liability nor provide warranty for any
 *   of this software. This material is provided "AS-IS" in the hope
 *   that it may be useful for others.
 *   
 **/

#define _GNU_SOURCE

#include <string.h>
#include <stdio.h>
#include <expat.h>
#include <my_stdio.h>

#include "xml_proc.h"

#ifdef XML_LARGE_SIZE
#  if defined(XML_USE_MSC_EXTENSIONS) && _MSC_VER < 1400
#    define XML_FMT_INT_MOD "I64"
#  else
#    define XML_FMT_INT_MOD "ll"
#  endif
#else
#  define XML_FMT_INT_MOD "l"
typedef unsigned long XML_Size; /* for expat-1.95.7-4 compatibility */
#endif


#define BUFFSIZE 8192


static void print_start_element(void *parser, const char *name,
				const char **atts)
{
	int i;
	FILE *fp = XMLP_GET_FPO(parser);

	if (XMLP_IF_OUT_FLAG(parser)) {
		fprintf(fp, "<%s", name);		
		for (i=0; atts[i]; i+=2) {
			fprintf(fp, " %s=\"%s\"", atts[i], atts[i+1]);
		}
		fprintf(fp, ">");
	}

	return;
}


static int tag_is(const char *tag, const char *name)
{
	int len = strlen(name);
	
	if (strlen(tag) == len && strncmp(tag, name, len) == 0)
		return 1;

	return 0;
}


static void XMLCALL start_element(void *parser, const char *name,
				  const char **atts) 
{
	int i;
	FILE *fp =  XMLP_GET_FPO(parser);
	struct blob *bl_id;

	if (tag_is(name, "rawdata")) {
		bl_id = blob_set_rawdata(XMLP_GET_BLOB(parser), atts);
		BLOB_GET_SLICE_REFID(bl_id) = XMLP_GET_CURRENT_REFID(parser);
		BLOB_GET_POSANGLE(bl_id) = XMLP_GET_CURRENT_POSANGLE(parser);
		if (XMLP_GET_PROC_RAWDATA(parser) != NULL) {
			XMLP_GET_PROC_RAWDATA(parser)
				(bl_id, XMLP_GET_PROC_PARAMETERS(parser));
		}
		//printf("# POSANGLE: %f\n", BLOB_GET_POSANGLE(bl_id));
		if (XMLP_IF_OUT_FLAG(parser) ) {
			fprintf(fp, "<%s", name);
			fprintf(fp," blobid=\"%d\" rays=\"%d\" type=\"%s\" bins=\"%d\" min=\"%.2f\" max=\"%.2f\" depth=\"%d\"",
				BLOB_GET_ID(bl_id), BLOB_GET_RAYS(bl_id), 
				BLOB_GET_TYPE_STR(bl_id), BLOB_GET_BINS(bl_id), 
				BLOB_GET_MIN(bl_id), BLOB_GET_MAX(bl_id), 
				BLOB_GET_DEPTH(bl_id) );	
			fprintf(fp, ">");
		}
	}else if (tag_is(name, "rayinfo")) {
		bl_id = blob_set_rayinfo(XMLP_GET_BLOB(parser), atts);
		BLOB_GET_SLICE_REFID(bl_id) = XMLP_GET_CURRENT_REFID(parser);
		BLOB_GET_POSANGLE(bl_id) = XMLP_GET_CURRENT_POSANGLE(parser);
		//PDBG("# POSANGLE: %f", BLOB_GET_POSANGLE(bl_id));
		print_start_element(parser, name, atts);
		} else if (tag_is(name, "history")) {
		XMLP_SET_OUT_FLAG(parser, 0);
	}else if (tag_is(name, "radarinfo")) {
		if (XMLP_GET_RADAR_INFO(parser) == NULL)
			XMLP_SET_RADAR_INFO(parser, ri_make());
		ri_set_pos_data(XMLP_GET_RADAR_INFO(parser), atts);
		print_start_element(parser, name, atts);

	}else if (tag_is(name, "scan")) {
		if (XMLP_GET_RADAR_INFO(parser) == NULL)
			XMLP_SET_RADAR_INFO(parser, ri_make());
		ri_set_time_data(XMLP_GET_RADAR_INFO(parser), atts);
		print_start_element(parser, name, atts);
	} else if (tag_is(name, "slice")) {
		for (i=0; atts[i] ; i+=2) {
			if ( strncmp(atts[i], "refid", 5) == 0 )
				XMLP_GET_CURRENT_REFID(parser) = 
					atoi(atts[i+1]);
		}
		print_start_element(parser, name, atts);
	} else if (tag_is(name, "posangle")) {
		XMLP_GET_STORE_TYPE(parser) = STORE_DOUBLE;
		XMLP_GET_STORE_VALUE(parser) = 
			&XMLP_GET_CURRENT_POSANGLE(parser);
		print_start_element(parser, name, atts);
	} else if (tag_is(name, "stoprange")) {
		XMLP_GET_STORE_TYPE(parser) = STORE_INT;
		XMLP_GET_STORE_VALUE(parser) = 
			&XMLP_GET_STOPRANGE(parser);
		print_start_element(parser, name, atts);
	} else if (tag_is(name, "beamwidth")) {
		XMLP_GET_STORE_TYPE(parser) = STORE_DOUBLE;
		XMLP_GET_STORE_VALUE(parser) = 
			&XMLP_GET_BEAMWIDTH(parser);
		print_start_element(parser, name, atts);
	} else  {
		print_start_element(parser, name, atts);
	}
}


static void XMLCALL end_element(void *parser, const char *name) {

	if (tag_is(name, "history")) {
		XMLP_SET_OUT_FLAG(parser, 1);
	} else {
		if (XMLP_IF_OUT_FLAG(parser))
			fprintf(XMLP_GET_FPO(parser), "</%s>", name);
	}

	return;
}


static void XMLCALL characters(void *parser, const char *ch, int len) {
	int i;
	char val[len+1];

	if (XMLP_IF_OUT_FLAG(parser)) {
		for (i=0; i < len; i++) 
			fprintf(XMLP_GET_FPO(parser), "%c", ch[i]);
	}
	if (XMLP_GET_STORE_VALUE(parser) != NULL) {
		snprintf(val, len+1, "%s",ch);
		switch(XMLP_GET_STORE_TYPE(parser)) {
		case STORE_DOUBLE:
			*((double *)XMLP_GET_STORE_VALUE(parser)) = atof(val);
			break;
		case STORE_INT:
			*((int *)XMLP_GET_STORE_VALUE(parser)) = atoi(val);
			break;
#if 0
		case STORE_CHAR:
			memcpy((char *)XMLP_GET_STORE_VALUE(parser), val, len);
			break;
#endif
		default:
			PVERB(PERR,"Not implemente STORE_TYPE");
		}
		XMLP_GET_STORE_VALUE(parser) = NULL;
		XMLP_GET_STORE_TYPE(parser) = STORE_UNKNOW;
	}
		
}


XML_Parser xmlp_init(void)
{
	XML_Parser parser = XML_ParserCreate(NULL);
	xmlp_udata *data = (xmlp_udata *) malloc(sizeof(xmlp_udata));

	XML_SetUserData(parser, data);
	XML_UseParserAsHandlerArg(parser);
	XML_SetElementHandler(parser, start_element, end_element);
	XML_SetCharacterDataHandler(parser, characters);

	XMLP_SET_BLOB(parser, NULL);
	XMLP_SET_FPO(parser, NULL);
	XMLP_SET_PROC_PARAMETERS(parser, NULL);
	XMLP_SET_PROC_RAWDATA(parser, NULL);
	XMLP_SET_RADAR_INFO(parser, NULL);
	XMLP_SET_INIVIT_OUT_FLAG(parser, 0);
	XMLP_GET_STORE_VALUE(parser) = NULL;
	XMLP_GET_ADD_OUT_FILE(parser) = 0;

	return parser;
}


int xmlp_set_FILE_out(XML_Parser parser, FILE *fp)
{
	xmlp_udata *data = (xmlp_udata *) XML_GetUserData(parser);

	if( data != NULL) {
		data->fpo = fp;
		return 0;
	}

	return -1;
}


int xmlp_open_file_out(XML_Parser parser, char *fname)
{
	xmlp_udata *data = (xmlp_udata *) XML_GetUserData(parser);
	FILE *fp;
	char *file_mode[] = {"w","a"};

	if( data != NULL) {
		if ((fp = fopen(fname, 
				file_mode[XMLP_GET_ADD_OUT_FILE(parser)])) == 
		    NULL) {
			PVERB(PERR,"Can't open file: %s", fname);
			return -1;
		}	
		data->fpo = fp;
	}

	return 0;
}


FILE *xmlp_get_FILE_out(XML_Parser parser)
{
	return XMLP_GET_FPO(parser);
}


int xmlp_close_file_out(XML_Parser parser)
{
	return fclose(XMLP_GET_FPO(parser));
}


void xmlp_free(XML_Parser parser)
{	
	ri_free(XMLP_GET_RADAR_INFO(parser));
	free(((xmlp_udata *) XML_GetUserData(parser)));
	XML_ParserFree(parser);
}


int xmlp_run_on(XML_Parser parser, char *fname) 
{
	FILE *fp;
	char buff[BUFFSIZE];
	int done = 0;

	if ((fp = fopen(fname, "r"))  == NULL) {
		PVERB(PERR,"Can't open file: %s", fname);
		return -1;
	}

	XMLP_SET_OUT_FLAG(parser, 1);
	do {
		int len  = fread(buff, 1, BUFFSIZE, fp);
		done = feof(fp);
		if (XML_Parse(parser, buff, len, done) == XML_STATUS_ERROR) {
			PDBG("%s at line %" XML_FMT_INT_MOD "u",
			     XML_ErrorString(XML_GetErrorCode(parser)),
			     (XML_Size) XML_GetCurrentLineNumber(parser));
			done = 1;
		}

	} while (!done);
	if( XMLP_IF_OUT_FLAG(parser) )
		fprintf(XMLP_GET_FPO(parser),"\n");
	fclose(fp);
	
	return 0;
}


